// -*- C++ -*-
/*!
 * @file  SURFObjectFinder.cpp
 * @brief Object Finder from Image using SURF
 * @date $Date: 2010-04-08 16:18:32 +0900 (木, 08  4月 2010) $
 *
 * $Id: SURFObjectFinder.cpp 338 2010-04-08 07:18:32Z eus $
 */

#include "SURFObjectFinder.h"

// Module specification
// <rtc-template block="module_spec">
static const char* surfobjectfinder_spec[] =
  {
    "implementation_id", "SURFObjectFinder",
    "type_name",         "SURFObjectFinder",
    "description",       "Object Finder from Image using SURF",
    "version",           "1.0.0",
    "vendor",            "Hiroaki Yaguchi JSK",
    "category",          "3D Estimation",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    "exec_cxt.periodic.rate", "1.0",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
SURFObjectFinder::SURFObjectFinder(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_InImageIn("InImage", m_InImage),
    m_ObjectPoseOut("ObjectPose", m_ObjectPose),
    m_ObjectIndexOut("ObjectIndex", m_ObjectIndex),

    // </rtc-template>
	dummy(0)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("InImage", m_InImageIn);
  
  // Set OutPort buffer
  registerOutPort("ObjectPose", m_ObjectPoseOut);
  registerOutPort("ObjectIndex", m_ObjectIndexOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  surf=new SURFConverter();
  surfman=new SURFManager(surf);
  inimg=NULL;
  outimg=NULL;
  surf_result=-1;
  isverbose=true;

  cvNamedWindow("inimage");
  cvNamedWindow("outimage");
  cvWaitKey(10);

  //

  string dir_path=".";
  string ini_file="SURFObjectFinder.ini";
  readIni(dir_path,ini_file);

  surfman->setMatchedNumThres(40);
}

/*!
 * @brief destructor
 */
SURFObjectFinder::~SURFObjectFinder()
{
  delete surfman;
  delete surf;

  cvDestroyWindow("inimage");
  cvDestroyWindow("outimage");
}



RTC::ReturnCode_t SURFObjectFinder::onInitialize()
{
  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t SURFObjectFinder::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SURFObjectFinder::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SURFObjectFinder::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SURFObjectFinder::onActivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SURFObjectFinder::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t SURFObjectFinder::onExecute(RTC::UniqueId ec_id)
{
  int width,height,depth;
  int idx;

  if(m_InImageIn.isNew()){
    m_InImageIn.read();

    width=m_InImage.width;
    height=m_InImage.height;
    depth=m_InImage.bpp/8;

    cerr << "recieved image: " << width << " x " << height << " x " << depth  << endl;

    // convert TimedImage to IplImage 
    Mat img;

    if(convertToCV(m_InImage,img)){
      IplImage iplimg=img;
      inimg=cvCloneImage(&iplimg);
    }else{
      cerr << "convert failed." <<endl;
      return RTC::RTC_OK;
    }

    //
    
    if(outimg!=NULL) cvReleaseImage(&outimg);
    outimg=cvCloneImage(inimg);

    findObject();

    m_ObjectIndex.data=surf_result;
    m_ObjectPose.data.length(6);
    for(int i=0;i<3;i++){
      m_ObjectPose.data[i]=pos[i];
      m_ObjectPose.data[i+3]=aangle*aaxis[i];
    }

    m_ObjectIndexOut.write();
    m_ObjectPoseOut.write();
  }

  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t SURFObjectFinder::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SURFObjectFinder::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SURFObjectFinder::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SURFObjectFinder::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SURFObjectFinder::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

int SURFObjectFinder::findObject(){
  //int ret=surfman->proc(inimg);
  //int ret=surfman->procList(inimg);
  surf_result=surfman->procList(inimg);

  fprintf(stderr,"ret = %d\n",surf_result);

  //if(ret==1){
  //if(ret>=0){
  if(surf_result>=0){
    fprintf(stderr,"draw features\n");
    int fnum=surfman->getFeatureNum();
    float *features_ser=surfman->getSerFeatures();
    for(int i=0;i<fnum;i++){
      cvCircle(outimg,cvPoint((int)features_ser[i*2],(int)features_ser[i*2+1]),3,CV_RGB(0,255,0),1);
    }

    fprintf(stderr,"draw rectangle\n");
    float *corners=surfman->getCorners();
    cvLine(outimg,cvPoint((int)corners[0],(int)corners[1]),cvPoint((int)corners[2],(int)corners[3]),CV_RGB(255,0,0),3);
    cvLine(outimg,cvPoint((int)corners[2],(int)corners[3]),cvPoint((int)corners[4],(int)corners[5]),CV_RGB(255,0,0),3);
    cvLine(outimg,cvPoint((int)corners[4],(int)corners[5]),cvPoint((int)corners[6],(int)corners[7]),CV_RGB(255,0,0),3);
    cvLine(outimg,cvPoint((int)corners[6],(int)corners[7]),cvPoint((int)corners[0],(int)corners[1]),CV_RGB(255,0,0),3);

    fprintf(stderr,"create transform broadcast\n");
    float *params=surfman->getMarkParam();

    aangle=sqrt(params[3]*params[3]+params[4]*params[4]+params[5]*params[5]);
    for(int i=0;i<3;i++){
      pos[i]=params[i]*0.001;
      aaxis[i]=params[i+3]/aangle;
    }
  }
  if(isverbose){
    cvShowImage("outimage",outimg);
    cvShowImage("inimage",inimg);
    cvWaitKey(10);
  }
  //return ret;
  return surf_result;
}

bool SURFObjectFinder::readIni(string &path,string &file){
  string buf,str0,str1;
  vector<string> imagefiles;
  vector<float> widthlist,heightlist;
  float wmm,hmm;

  string path2;
  string fullpath;

  if(path==""){
    path2=path;
  }else{
    path2=path+"/";
  }

  fullpath=path2+file;

  ifstream ifs(fullpath.c_str());
  if(!ifs){
    fprintf(stderr,"ini file %s not found.\n",fullpath.c_str());
    return false;
  }

  while(getline(ifs,buf)){
    stringstream sin(buf);
    sin>>str0;

    if(str0=="image"){
      sin >> str1;
      imagefiles.push_back(str1);

      sin >> str1;
      try{
        wmm=boost::lexical_cast<float>(str1);
      }catch(boost::bad_lexical_cast &ex){
        wmm=-1.0;
      }
      sin >> str1;
      try{
        hmm=boost::lexical_cast<float>(str1);
      }catch(boost::bad_lexical_cast &ex){
        hmm=-1.0;
      }
      widthlist.push_back(wmm);
      heightlist.push_back(hmm);
    }
  }

  //

  if(imagefiles.size()<1){
    fprintf(stderr,"no image file registrated.\n");
    return false;
  }

  IplImage *img;

  // add first image
  fullpath=path2+imagefiles[0];
  img=cvLoadImage(fullpath.c_str(),CV_LOAD_IMAGE_ANYDEPTH | CV_LOAD_IMAGE_ANYCOLOR);
  if(img==NULL){
    fprintf(stderr,"%s: not found!\n",imagefiles[0].c_str());
    return false;
  }
  fprintf(stderr,"%s: read success!\n",imagefiles[0].c_str());
  surfman->setRefImage(img);
  if(widthlist[0]>0.0 && heightlist[0]>0.0)
    surf->pushRefImageList(img,widthlist[0],heightlist[0]);
  else
    surf->pushRefImageList(img);
  cvReleaseImage(&img);

  //
  for(int i=1;i<imagefiles.size();i++){
    fullpath=path2+imagefiles[i];
    img=cvLoadImage(fullpath.c_str(),CV_LOAD_IMAGE_ANYDEPTH | CV_LOAD_IMAGE_ANYCOLOR);
    if(img==NULL){
      fprintf(stderr,"%s: not found!\n",imagefiles[i].c_str());
      return false;
    }
    fprintf(stderr,"%s: read success!\n",imagefiles[i].c_str());
    if(widthlist[i]>0.0 && heightlist[i]>0.0)
      surf->pushRefImageList(img,widthlist[i],heightlist[i]);
    else
      surf->pushRefImageList(img);
    cvReleaseImage(&img);
  }

  surf->convertRefList();

  return true;
}



extern "C"
{
 
  void SURFObjectFinderInit(RTC::Manager* manager)
  {
    RTC::Properties profile(surfobjectfinder_spec);
    manager->registerFactory(profile,
                             RTC::Create<SURFObjectFinder>,
                             RTC::Delete<SURFObjectFinder>);
  }
  
};


